// Variables defined here are "Global" that is, accessible to every sub-program for reading and writing.
// We need to declare here some of the variable whose values will be defined in the setup() routine
// and later used in the draw() routine and its sub-programs.

boolean CNTRL;

Table histTab;

int xmax = 1900;              // Display canvas maximum width
int ymax = 1000;              // Display canvas maximum height

int nrmax0 = 560;             // If nrmax0 and nimax0 are non-zero then 
int nimax0 = 1000;            // these are the dimensions used for the canvas.
                              // If they are zero the xmax and ymax define maximum
                              // dimensions, but nrmax and nimax will be calculated to stay
                              // within these limits while preserving the original dimension ratio.

int count = -1;               // For counting the number of times draw() has been called.


float RadToDeg = 180.0/PI;    // Conversion factor from radians to degrees.
float DegToRad = PI/180.0;    // Conversion factor from degrees to radians

String Picture="MiamiLake.jpg";// Source image

int series=1;                 // Used to distinguish output file names from previous runs.

float dt      = 0.5;
float time    = 0.0;
float tmax    = 600.0;       // Time after which dynamic update stops


boolean dynamic = true;       // The parameters controlling the form of the conformal map change with time.
  

boolean redraw  = true;       // redraw == true ==> the draw() function regernates the canvas.
                              // redraw = false ==. draw() does nothing. 


boolean replay  = false;           // If this is true, then the file denoted by replayFile is used to define
                                   // a sequence of conformal mapping parameters, which are successively played
                                   // until the file runs out.
String  replayFile ="replay.csv";  // Nothing happens unless replay==true.

boolean autosave= false;            // Saves and output file at every cycle of draw().
                                   // Typically used with dynamic==true or replay==true


Figure epifig;                // This is the object that holds both the photograph and the modified output image.

class Vector{
  float x;
  float y;
  
  Vector(float xi,float yi)
  {
    x = xi;
    y = yi;
  }
}

void setup()                  // Executed once when Processing starts-up.
{
  size(100,100);              // Size of our visible canvas - has to have explicit integer parameters by Processing rules. //<>//
  surface.setResizable(true);

  epifig = new Figure(Picture,9,1,1);  // Create data structures to hold the modified image
  epifig.roll = 2;                     // Determines what occurs when we go off the edge of the colour  map.

  frameRate(1);               // Make this smaller if you want more time to see what is happening.
  
  println(width,height);
  
  epifig.defineReplay();

  
  if(replay){ epifig.replayDoc(replayFile); }

}


void draw()
{

  stroke(255,50);
  
  if(time > tmax) { redraw = false;}

  if(redraw)             // redraw gets reset to false if one of the harmonic amplitudes or phases changes. 
    {
    if(!dynamic){ redraw = false;}  // Once we have drawn the figure with new parameters, set to 0 to avoid redraws.

    if(replay) { epifig.replayDocNext();}
    
    count++;                 // Sometimes useful to know the number of times redraw has been executed.

    background(255);         // Make the background colour white - erasing everything
                             // from the last time draw() was executed.   

    epifig.drawFig(time);

    if(autosave) { epifig.saveSmallFig(); }   

    time = time + dt;
  }
}

void mouseReleased()
{
  if(mouseButton == LEFT)
    {
      println("Mouse position: ", mouseX,mouseY, " Fraction=",float(mouseX)/width,float(mouseY)/height);
    }
}

void keyReleased()
{
  // This is called by the Processing system once whenever a key is released.
  // The value of the key is stored in the "key" global variable.
  // We use the "released" event as a user signal for certain actions
  // because we can guarantee that it is a descrete event.
  // (In contrast the keyPressed() routine is called multiple times while a key is held down.)

  if(key == CODED)
    {
    if(keyCode == CONTROL)
        {
          CNTRL = false;
        }
      }  

  if(key == 's') {
    println("Save...");
    epifig.saveSmallFig();  // Save a copy of the current image as a JPEG file.
  }                         // The name of the time is constructed according to
                            // the exact time the event was actioned so multiple
                            // requests always generate new files.

  if(key == 'S') {
    println("Save...");
    epifig.saveLargeFig();  // Save a copy of the current image as a JPEG file.
  }  

}  

void keyPressed()
{ 
  int harm=11;  

  float xfrac = float(mouseX)/float(width);
  float yfrac = float(mouseY)/float(height);

  if(key == CODED)
    {
    if(keyCode == CONTROL)
        {
          CNTRL = true;
        }
     }
  
  if(key == '1') {if(CNTRL) { harm = -1;} else  { harm = 1; }} 
  if(key == '2') {if(CNTRL) { harm = -2;} else  { harm = 2; }} 
  if(key == '3') {if(CNTRL) { harm = -3;} else  { harm = 3; }}
  if(key == '4') {if(CNTRL) { harm = -4;} else  { harm = 4; }}
  if(key == '5') {if(CNTRL) { harm = -5;} else  { harm = 5; }} 
  if(key == '6') {if(CNTRL) { harm = -6;} else  { harm = 6; }}
  if(key == '7') {if(CNTRL) { harm = -7;} else  { harm = 7; }}
  if(key == '8') {if(CNTRL) { harm = -8;} else  { harm = 8; }}
  if(key == '9') {if(CNTRL) { harm = -9;} else  { harm = 9; }}
  if(key == '0') {if(CNTRL) { harm =  0;} else  { harm = 10;}}

  
  //println("Key ",key);
    
  if(harm >= -9 && harm <= 10)
    {       
     float amp   = xfrac;    // amp in the range 0.0 -> 1.0.
     float phase = yfrac;    // phase in the range 0.0 -> 1.0 (scales to TWO_PI)
     println("Harmonic ",harm, " amplitude=",amp, " phase=",phase);

     epifig.setHarmonic(harm,xfrac,yfrac); 
     redraw=true;
    }
    
  if(key == 'z' )         
    { 
      epifig.setHarmonic(-100,xfrac,yfrac); 
      redraw=true;
    }
    
  else if(key == 'r') 
    { 
    epifig.roll = (epifig.roll + 1)%3; 
    redraw = true;
    println("epifig.roll =",epifig.roll);
    }
    
  else if(key == 't') 
     { // Switch to dynamic/non-dynamic changes
     if(dynamic){dynamic=false;}
     else       {dynamic=true;}

     redraw = true;
     }
  
  else if(key == 'c')
    {
      epifig.setCentre(xfrac,yfrac);
      redraw = true;
    }

}


class Figure
{ // This class handles the modification of the original image.

  
  int nimax;                    // Will hold height of source colour map
  int nrmax;                    // Will hold width of source colour map
  
  float dratio;                 // Yheight/Xwidth
                                
  PImage cmap;                  // Current image - original photograph from disk.
  PImage omap;                  // High resolution output image  

  float xsc;                    // Drop falls here
  float ysc;                    // Drop falls here
  
  float xmc;                    // Used as the centre of the transformation formula - x coordinate
  float ymc;                    // Used as the centre of the transformation formula - y coordinate

  float rx=1.0;                 // Overall scale factor in the x direction.
  float ry=1.0;                 // Overall scale factor in the y direction.
  
  int hmax;                     // Number of harmonics allowed (+/-).  Arrays below are dimensioned [2*hmax+1]
                                // so r[0]      is amplitude of harmonic -h
                                //    r[hmax]   is amplitude of harmonic 0
                                //    r[hmax+1] is amplitude of harmonic 1
                                //    r[2*hmax] is amplitude of harmonic h

  float[] r;                    // The amplitude of radial displacement
  float[] t;                    // The amplitude of theta displacement
  float[] p;                    // Phases of each harmonic;
  float[] f;                    // The relative frequency of each epicycle;
  float[] v;                    // propagation velocity at this frequency;
  float[] dc;                   // Decay constant for this frequency,


  String photoRoot;             // Root of the name of the original photograph.
  
  int saveCount = 0;            // Number of times a save operation is requested.
  int redrawCount=0;            // Number of time a redraw is requested.
  int lastSaveAt =0;            // The "redrawCount" value at which a save was last executed.
                                // This is used to give the same filename to "small" and "large" output images.
                                // when large-save is requested after a small-save.

  int roll  = 0;                // Controls how boundaries are treated.
  
  Table replayTable;            // Read from disk containing a list of previous image configurations.
  int   replayRow = -1;         // Used to iterate through the rows in replayTable.
  

  Figure(String photoName, int nharm, int nsymm, int symstart)
    {
    
    photoRoot = photoName.substring(0,photoName.length()-4);
    
    cmap = loadImage(photoName); // Load the original image
    cmap.loadPixels();           // Make its pixels addressable in the cmap.pixels[] array.

    println("Photo: width,height: ",cmap.width, cmap.height);    
    // Find division factor for image that brings size down to less than screen size

    if(nrmax0 == 0 && nimax0 == 0)
        {
        for(int id=1;id<10;id++)
          {
            nrmax = cmap.width/id;
            nimax = cmap.height/id;
            
            println("id=",id," nrmax=",nrmax," nimax=",nimax);
            if( nrmax <= xmax && nimax <= ymax)
              {
                break;
              }
          }
        }
    else
        {
          nrmax = nrmax0;
          nimax = nimax0;
        }

    dratio = float(nimax)/float(nrmax);
    
    println("nrmax,nimax: ",nrmax,nimax);
    surface.setSize(nrmax, nimax);

    omap = createImage(cmap.width*5/5, cmap.height*5/5, RGB);
    omap.loadPixels();
    
    hmax   = nharm;
       
    r = new float[hmax];        // Array of harmonic amplitudes
    t = new float[hmax];        // Array of amplitudes for theta variation
    p = new float[hmax];        // Array of harmonic phases
    f = new float[hmax];        // The harmonic frequencies.  
    v = new float[hmax];
    dc= new float[hmax];
    
    for(int h=0;h<hmax;h++)
      {
      r[h]  = 0.0;             // Need to ensure that all circle radii are initialised - set all to zero to start with.
      t[h]  = 0.0;
      p[h]  = 0.0;             // Phases all zero.
      f[h]  = 1.0*h;           // Set frequencies to be n=k*mod(m) to give (m-1)-fold figure symmetry.
      v[h]  = 0.01 + 0.002*h;
      dc[h] = 20.0;
      }

    r[1] = 0.3;                       // The first harmonic amplitude. 
    r[2] = 0.15;
    r[3] = 0.12;
    r[4] = 0.01;
    t[1] = 0.5;
    t[2] = 0.2;    

    rx = 1.0;                         // The fundamental size scale - x direction;
    ry = 1.0;                         // The fundamental size scale - y direction;
   
    xsc = 0.5;                  
    ysc = 0.5;
    xmc = 0.0;
    ymc = 0.0;   
        
    for(int h=0;h<hmax;h++)
      {
      if(r[h]>0.0){
         println( h,"harm[",h,"]=( r=",r[h],"p=",p[h],")" );
        }
      }
    println("rx,ry=",ry,ry," xsc,xsy=",xsc,ysc," xmc,ymc=",xmc,ymc);
}

  
  void setHarmonic(int ih, float xfrac, float yfrac)
  { 
    // Define value for a harmonic component
    //   xfrac : 0.0 <= xfrac <= 1.0
    //   yfrac : 0.0 <= yfrac <= 1.0
    //
    //   amp   : float  amp   >= 0.0 && amp   <= 1.0.
    //   phase : float  phase >= 0.0 && phase <= TWO_PI.
    //
    // ih == 0   ==> rx  = 10.0*xfrac; ry=10.0*yfrac;
    // ih >= 1   ==> r[ih] = amp && p[ih] = phase.
    //
    //---------------------------------------------------

    float amp   = xfrac;
    float phase = TWO_PI*yfrac;
    
    if(ih > 0 && ih <= hmax)
      { 
        r[ih-1] = amp; 
        p[ih-1] = phase;
      }
      
    else if(ih == 10)       // corresponds to the '0' key being pressed.          
        { 
        rx  = 10.0*amp;    
        ry  = 10.0*amp;
      }
    
    else if(ih == -100)
      {
    
      for(int h=0;h<hmax;h++)
        {
        r[h] = 0.0;           // Need to ensure that all circle radii are initialised - set all to zero to start with.
        t[h] = 0.0;
        p[h] = 0.0;           // Phases all zero.
        }
            
      r[1] = 0.1;                       // The first harmonic amplitude. 
      t[1] = 0.0;
      t[2] = 0.0;
      //r[2] = 0.07;
      //r[3] = 0.05;
      //r[4] = 0.03;
    
      rx = 1.0;                         // The fundamental size scale - x direction;
      ry = 1.0;                         // The fundamental size scale - y direction;
     
      xsc = 0.5;                  
      ysc = 0.5;
      xmc = 0.0;
      ymc = 0.0;  
      }
    
    for(int h=0;h<hmax;h++)
      {
      if( r[h]>0.0 ){
         println( h,"harm[",h,"]=( r=",r[h],"p=",p[h],")" );
        }
      }
    println("rx,ry=",ry,ry," xsc,xsy=",xsc,ysc," xmc,ymc=",xmc,ymc);
  }
  
  void setCentre(float xfrac, float yfrac)
  {
    // Define the centre point of the transformation in the source plane
    // in terms of the fraction of width and height denoted by xfrac and yfrac.
    xmc = 2.0*(xfrac-xsc);
    ymc = 2.0*(yfrac-ysc);
    
    println("xmc=",xmc," ymc=",ymc);
  }
    
  
  void drawFig(float time)
  { // Calculate and draw the transformed figure.
    //--------------------------------------------------------------
    
    //println( "drawFig"); //<>//
    redrawCount++;
    
    if(dynamic)
      {
      println("Dynamic time: ",time);
      redraw = true;  
      }
    
    loadPixels();                            // Access the screen pixels
    transform(cmap, width,height, pixels);   // Calculate output on the canvas pixel array.
    updatePixels();                          // update the screen.

  }
  
  
  void transform(PImage imap, int mwidth, int mheight,int[] outPixels)
  {

    for(int nr=1; nr<=mwidth; nr++)                          // This iterates over the real axis of the target plane
      {
      float x = (xsc-0.5) -1.0 + 2.0*float(nr-1)/float(mwidth-1);      // xsc=0.5 && nr==1      ==> x == -1.0 
                                                                       // xsc=0.5 && nr==mwidth ==> x ==  1.0

      for(int ni=1; ni <= mheight; ni++)                     // This iterates over the imaginary axis of the target plane.
        {
        float y =((ysc-0.5) -1.0 + 2.0*float(ni-1)/float(mheight-1));  // ysc=0.5 && ni==1       ==> y == -1.0 
                                                                       // ysc=0.5 && ni==mheight ==> y ==  1.0

        Vector d1 = displacement(time-1.0,  x,y, xmc-0.1,ymc-0.2);        
        Vector d2 = displacement(time-20.0, x,y, xmc-0.3,ymc+0.6); 
        Vector d3 = displacement(time-40.0, x,y, xmc+0.4,ymc+0.3);
        Vector d4 = displacement(time-65.0, x,y, xmc+0.3,ymc-0.4);

        float x2 = x + d1.x+d2.x+d3.x+d4.x;
        float y2 = y + d1.y+d2.y+d3.y+d4.y;

        color c = getColour(imap,x2,y2);             // Find a colour on the source colour map.

        outPixels[(ni-1)*mwidth + nr-1] = c;     

        }
      }  
  }
  
  Vector displacement(float time,float x,float y,float xmc,float ymc)
  { 
    // Work out displacement due to wave train.
    // x   horizontal coordinate -1<=x<=+1 over image
    // y   horizontal coordinate -1<=y<=+1 over image
    //-------------------------------------------------------------------------
    
    float y1 = (y-ymc)*dratio;                           // Displacement from wave train centre,
    float x1 = x-xmc;                                    // with x1,y1 having same scale on the screen.
                
    float theta = atan2(y1,x1);                          // Switch to polar coordinates, angle of vector to wave centre.
    float rad   = sqrt(x1*x1 + y1*y1);                   // and radius from the centre of the wave train.
        
    float dr = dRad  (rad,theta,time);
    float dth= dTheta(rad,theta,time);
    
    Vector d = new Vector(
                        dr*cos(theta) + rad*dth*sin(theta),
                       (dr*sin(theta) - rad*dth*cos(theta))/dratio);        

    return d;
  }
 
  float dRad(float rad,float th,float time)
  {
    // rad varies from 0 upwards
    // th varies from 0 to TWO_PI
    //--------------------------------------------------------------

    float dr = 0.0;
    
    for(int h=0;h<hmax;h++)
        {        
        if(r[h] > 0.0 && v[h] > 0.0 && time*v[h] > rad)
          {           
          dr = dr + exp(-(time-rad/v[h])/dc[h])* r[h] * sin((f[h] + time-rad/v[h])/TWO_PI);
          }
      }
 
    return dr;
  }
  
  float dTheta(float rad,float th,float time)
  {
    // rad varies from 0 upwards
    // th varies from 0 to TWO_PI
    //--------------------------------------------------------------        

    float dth = 0.0;
    
    for(int h=0;h<hmax;h++)
        {        
        if(t[h] > 0.0 && v[h] > 0.0 && time*v[h] > rad)
          {           
          dth = dth + exp(-time/dc[h])*t[h]*cos((f[h] + time-rad/v[h])/TWO_PI);
          }
      }
      return dth;   // Does nothing.
  }
  
  color getColour(PImage map, float x,float y) 
  { 
    // Map an X,Y coordinate pair to a colour
    // - expect x and y to run from -1 to +1.
    //=======================================================
  
     int ix1 = int( (float)map.width *(x/2.0 + 0.5) ); 
     int iy1 = int( (float)map.height*(y/2.0 + 0.5) );
     
     int ix2=ix1;
     int iy2=iy1;
     
     int ix=ix1;
     int iy=iy1;
     
     int pixel=0;
     
     if(roll == 0)
       { // Truncate the coordinates at (0,map.width-1) or (0, map.height-1)
       
       if     (ix1 > map.width -1) { ix2 = map.width -1; } 
       else if(ix1 < 0           ) { ix2 = 0;            }
       else                        { ix2 =ix1;           }

       if     (iy1 > map.height-1) { iy2 = map.height-1; }  
       else if(iy1 < 0           ) { iy2 = 0;            }
       else                        { iy2 = iy1;          }
       
       ix = ix2;
       iy = iy2;
       
       pixel = map.width*iy + ix;
       }
       
     else if(roll == 1)
       {  // Attach x=0 to x=map.width and y=0 to y=map.height.
         
       ix2 = ix1 % map.width;   // If x>=0 this is in range 0...map.width-1.
                                // If x<0  this is in the range -1 to -(map.width-1)
       iy2 = iy1 % map.height;  // If y>=0 this is in range 0...map.height-1.  
                                // If y<0  this is in the range -1 to -(map.height-1)
                                
       if(ix2 < 0) { ix = map.width -1 +ix2;}
       else        { ix = ix2;              }
       
       if(iy2 < 0) { iy = map.height-1 +iy2;}
       else        { iy = iy2;              }

       pixel = map.width*iy + ix;         
       }
       
     else if(roll == 2)
       {  
       
       ix2 = ix1 % map.width; 
       iy2 = iy1 % map.height;
       
       if     (ix2 < 0)            { ix = -ix2;}
       else if(ix1 > map.width-1 ) { ix =  map.width-1  - ix2; }
       else                        { ix =  ix2;}
     
       if(iy2 < 0)                 { iy = -iy2;}
       else if(iy1 > map.height-1) { iy =  map.height-1 - iy2; }
       else                        { iy =  iy2;}

       pixel = map.width*iy + ix;         
       }
     else
       {
       
       if     (ix1 > map.width -1) { ix2 = -1; } 
       else if(ix1 < 0           ) { ix2 = -1; }
       else                        { ix2 =ix1; }

       if     (iy1 > map.height-1) { iy2 = -1; }  
       else if(iy1 < 0           ) { iy2 = -1; }
       else                        { iy2 = iy1;}
       
       ix = ix2;
       iy = iy2;
       
       if(ix >=0&& iy >=0)
         {
         pixel = map.width*iy + ix;
         }
       else
         {
         pixel = -1;
         }
       }
           

     if( pixel >= 0 && pixel < map.width*map.height)
       {
       return map.pixels[pixel];
       }
     else
       {
       return color(0);
       }
  }  
   
  String dateStamp()
  {
      String syear   = String.valueOf( year()   );
      String smonth  = String.valueOf( month()  );
      String sday    = String.valueOf( day()    );
      String shour   = String.valueOf( hour()   );
      String sminute = String.valueOf( minute() );
      String ssecond = String.valueOf( second() );
    
      return syear+smonth+sday+"_"+shour+":"+sminute+":"+ssecond;  
  }
  
  void replayDoc(String fName)
  {      
    Table replayTable = loadTable(fName,"header"); 
    replayRow = -1;
  }
  
  void replayDocNext()
  {
    replayRow++;
    
    TableRow row = replayTable.getRow(replayRow);
    
    rx  = row.getFloat("rx");
    ry  = row.getFloat("ry");
    xsc = row.getFloat("xsc");
    ysc = row.getFloat("ysc");
    xmc = row.getFloat("xmc");
    ymc = row.getFloat("ymc");   
  
    for(int h= 0; h<hmax; h++)
      {
      r[h] = row.getFloat("r"+String.valueOf(h));
      p[h] = row.getFloat("p"+String.valueOf(h));
      p[h] = row.getFloat("v"+String.valueOf(h));      
      p[h] = row.getFloat("dc"+String.valueOf(h));      
      }
    redraw = true;         

  }

  void defineReplay()
  {
      
    histTab = new Table();
    
    histTab.addColumn("rx");
    histTab.addColumn("ry");
    histTab.addColumn("xsc");
    histTab.addColumn("ysc");
    histTab.addColumn("xmc");  
    histTab.addColumn("ymc");  
    histTab.addColumn("fName");
    
    for(int h=0; h<epifig.hmax; h++)
      {
      histTab.addColumn("r"+String.valueOf(h));
      histTab.addColumn("p"+String.valueOf(h));
      histTab.addColumn("v"+String.valueOf(h));    
      histTab.addColumn("dc"+String.valueOf(h));
      }
  }
  
  void saveDoc(String fName)
  {
    String cudoc      = photoRoot+"_"+String.valueOf(series)+"D.csv";   
    String fdoc       = photoRoot+"_"+String.valueOf(series)+"_"+String.valueOf(saveCount)+"D.csv";   // For documenting the output files
      
    TableRow histrow = histTab.addRow();

 
    histrow.setFloat("rx",rx);
    histrow.setFloat("ry",ry);
    histrow.setFloat("xsc",xsc);
    histrow.setFloat("ysc",ysc);
    histrow.setFloat("xmc",xmc);  
    histrow.setFloat("ymc",ymc);  
    histrow.setString("fName",fName);  

    for(int h= 0; h<hmax; h++)
      {
      histrow.setFloat("r"+String.valueOf(h),r[h]);
      histrow.setFloat("p"+String.valueOf(h),p[h]);
      histrow.setFloat("v"+String.valueOf(h),v[h]);
      histrow.setFloat("dc"+String.valueOf(h),dc[h]);
      }
  
    saveTable(histTab,cudoc);
  }
  
  void saveSmallFig()
  {

    if(lastSaveAt < redrawCount) {saveCount++;}

    lastSaveAt = redrawCount;
    
    String fNameSmall = photoRoot+"_"+String.valueOf(series)+"_"+String.valueOf(saveCount)+"S.jpg";
    saveDoc(fNameSmall);
        
    println("Saving...");
    saveFrame(fNameSmall);

  }
  
  void saveLargeFig()
  {
    // Save small and large copies of the current image
    //------------------------------------------------------------------------------------------    
    println("Save large image...");

    if(lastSaveAt < redrawCount) {saveCount++;}

    lastSaveAt = redrawCount;

    String fNameLarge = photoRoot+"_"+String.valueOf(series)+"_"+String.valueOf(saveCount)+"L.jpg";
    String fNameSmall = photoRoot+"_"+String.valueOf(series)+"_"+String.valueOf(saveCount)+"S.jpg";
    saveFrame(fNameSmall);
    saveDoc(fNameLarge);

    surface.setSize(omap.width,omap.height);
    println("Revised Width=",width," Height=",height);    
    println(omap.width, omap.height);
    
    omap.loadPixels();
    transform(cmap, omap.width,omap.height,omap.pixels);

    println("Transformed to large canvas....");
    omap.updatePixels();
    
    println("Saving higher-res image to file ",fNameLarge);

    omap.save(fNameLarge);
    println("Saved large image.... resizing to screen canvas size.");
    
    surface.setSize(nrmax,nimax);
    loadPixels();
    transform(cmap, width,height,pixels);  
    updatePixels();
    redraw=true;

  }
  //<>//
}